/*
 * Generation of server dispatch and handle functions.
 *
 * Copyright (C) 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include "common.h"
#include "dispatch.h"
#include "templates.h"



/* Generate dispatch function signatures. */

void write_dispatcher_signature(const char *name, enum signature_role role,
                                FILE *fp)
{
  fprintf(fp, dispatch_function_signature, name, name);
  fputs(get_signature_terminator(role), fp);
}

/* Generate handler function signatures. */

void write_handler_signature(const char *name, enum signature_role role,
                             FILE *fp)
{
  fprintf(fp, handle_function_signature, name, name);
  fputs(get_signature_terminator(role), fp);
}

/* Generate a dispatch function for the different operations. */

void write_dispatcher(struct signature *sig, FILE *fp, struct interface *iface)
{
  write_dispatcher_signature(iface->name, DEFINITION_ROLE, fp);

  /* Interpret an operation indicator in the word data if a protocol applies to
     the entire interface. */

  if (get_protocol(iface->attributes))
    fputs("  switch (ipc_message_get_word(msg, 0))\n  {\n", fp);
  else
    fputs("  switch (l4_msgtag_label(msg->tag))\n  {\n", fp);

  write_dispatcher_cases(sig, fp, iface, 0);
  fputs(server_function_dispatcher_body_epilogue, fp);

  fputs("}\n", fp);
}

/* Generate each dispatch possibility within an interface. */

void write_dispatcher_cases(struct signature *sig, FILE *fp,
                            struct interface *iface, int compound)
{
  char *opcode, *opname, *protocol, *prefix, *ref, *s;

  if (sig == NULL)
    return;

  opname = get_operation_name(iface, sig);
  opcode = get_opcode_identifier(NULL, opname);
  prefix = get_operation_wrapper_prefix(sig->attributes);
  ref = get_object_conversion(iface, compound);

  /* Generate a reply if appropriate. */

  if (have_attribute(sig->attributes, "completion"))
    s = dispatch_function_wrapper_case;
  else
    s = dispatch_function_reply_wrapper_case;

  /* Generate the case and invocation. */

  fprintf(fp, s, opcode, prefix, opname, ref);

  /* Generate the other cases. */

  write_dispatcher_cases(sig->tail, fp, iface, compound);

  /* Free allocated strings. */

  free(opcode);
  free(opname);
  free(ref);
}
